#include "Player.h"
#include "Engine.h"
#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtc/epsilon.hpp>
#include <limits>
#include <algorithm>
Player::Player(PathfindingNode* initNode, Model* model) : Renderable(initNode->pos, model) {
	lastNode = initNode;
	SetAnimation("stand");
}

void Player::Update() {
	if (!nodesToVisit.empty()) {
		// using 0.2 as epsilon because it may be dumb and keep going past ep range
		if (glm::all(glm::epsilonEqual(pos, nodesToVisit.back()->pos, 0.2f))) {
			lastNode = nodesToVisit.back();
			nodesToVisit.pop_back();
		}
		// if we're still not empty
		if (!nodesToVisit.empty()) {
			// if we have a node to go to and we haven't reached it yet, go towards it
			pos += (glm::normalize(nodesToVisit.back()->pos - pos) * Engine::Instance().GetDeltaTime()) * moveSpeed;
			// glm::lookAt is reversed (because it's for vertices to move in relation to camera)
			// so we have to invert it
			rotMat = glm::inverse(glm::lookAt(pos, nodesToVisit.back()->pos, glm::vec3(0.0f, 1.0f, 0.0f)));
		}
		else {
			SetAnimation("stand");
		}
	}
	UpdateAnimation();
}

void Player::MoveTowards(const glm::vec3& target) {
	nodesToVisit.clear(); // empty queue
	// store the node graph so we don't keep retrieving it
	std::vector<PathfindingNode*> nodegraph = Engine::Instance().GetActiveWorld().getNList();
	if (nodegraph.empty()) return;
	// this will store our destination node, for now leave it here
	PathfindingNode* dest = nodegraph[0];
	for (PathfindingNode* n : nodegraph) {
		// if this is the node we're coming from, its distance cost is 0
		if (n == lastNode) {
			n->dist = 0;
		}
		else {
			// if not, then its distance cost is infinity (i.e. a big number)
			n->dist = std::numeric_limits<int>::max();
		}
		// reset every node's parent and visited flag
		n->parent = nullptr;
		n->visited = false;
		// if this node is closer to raycast hit point than old one, select this one as tar
		if (glm::distance(n->pos, target) < glm::distance(dest->pos, target)) {
			dest = n;
		}
	}
	// now we can make all the nodes calculate dist and parenting
	// only if our target node is not the one we're already at!
	if (dest != lastNode) {
		// recursive function to calculate weightings and set parents
		lastNode->DoPath(dest);
		nodesToVisit.push_back(dest); // nodesToVisit is stored in reverse order, working back from dest
		PathfindingNode* par = dest;
		// walk up the parent chain until we reach the initial node (that has no parent
		// because we marked it as visited before starting)
		while ((par = par->parent) != nullptr) {
			// don't add the initial node to the list (we don't need to move to ourselves)
			if (par != lastNode) {
				nodesToVisit.push_back(par);
			}
		}
		SetAnimation("walk");
	}
}